# -*- coding: utf-8 -*-
"""
Celery signal handlers and custom Django signal.
"""

from __future__ import absolute_import, unicode_literals

import logging
from uuid import uuid4

from celery.signals import before_task_publish, task_failure, task_prerun, task_retry, task_success

from django.contrib.auth import get_user_model
from django.db import close_old_connections, transaction
from django.utils.module_loading import import_string

from user_tasks import user_task_stopped

from .exceptions import TaskCanceledException
from .models import UserTaskStatus
from .tasks import UserTaskMixin

LOGGER = logging.getLogger(__name__)


@before_task_publish.connect
def create_user_task(sender=None, body=None, **kwargs):  # pylint: disable=unused-argument
    """
    Create a :py:class:`UserTaskStatus` record for each :py:class:`UserTaskMixin`.

    Also creates a :py:class:`UserTaskStatus` for each chain, chord, or group containing
    the new :py:class:`UserTaskMixin`.
    """
    try:
        task_class = import_string(sender)
    except ImportError:
        return
    if issubclass(task_class.__class__, UserTaskMixin):
        arguments_dict = task_class.arguments_as_dict(*body['args'], **body['kwargs'])
        user_id = _get_user_id(arguments_dict)
        task_id = body['id']
        if body.get('callbacks', []):
            _create_chain_entry(user_id, task_id, task_class, body['args'], body['kwargs'], body['callbacks'])
            return
        if body.get('chord', None):
            _create_chord_entry(task_id, task_class, body, user_id)
            return
        parent = _get_or_create_group_parent(body, user_id)
        name = task_class.generate_name(arguments_dict)
        total_steps = task_class.calculate_total_steps(arguments_dict)
        UserTaskStatus.objects.get_or_create(
            task_id=task_id, defaults={'user_id': user_id, 'parent': parent, 'name': name, 'task_class': sender,
                                       'total_steps': total_steps})
        if parent:
            parent.increment_total_steps(total_steps)


def _create_chain_entry(user_id, task_id, task_class, args, kwargs, callbacks, parent=None):
    """
    Create and update status records for a new :py:class:`UserTaskMixin` in a Celery chain.
    """
    LOGGER.debug(task_class)
    if issubclass(task_class.__class__, UserTaskMixin):
        arguments_dict = task_class.arguments_as_dict(*args, **kwargs)
        name = task_class.generate_name(arguments_dict)
        total_steps = task_class.calculate_total_steps(arguments_dict)
        parent_name = kwargs.get('user_task_name', '')
        with transaction.atomic():
            if parent is None:
                # First task in the chain, create a status record for it
                parent = UserTaskStatus.objects.create(
                    is_container=True, name=parent_name, task_class='celery.chain', task_id=str(uuid4()),
                    total_steps=0, user_id=user_id)
            UserTaskStatus.objects.create(
                name=name, parent=parent, task_class=task_class, task_id=task_id, total_steps=total_steps,
                user_id=user_id)
            parent.increment_total_steps(total_steps)
            if parent_name and not parent.name:
                parent.set_name(parent_name)
    for callback in callbacks:
        links = callback.options.get('link', [])
        callback_class = import_string(callback.task)
        _create_chain_entry(user_id, callback.id, callback_class, callback.args, callback.kwargs, links, parent=parent)


def _create_chord_entry(task_id, task_class, message_body, user_id):
    """
    Create and update status records for a new :py:class:`UserTaskMixin` in a Celery chord.
    """
    args = message_body['args']
    kwargs = message_body['kwargs']
    arguments_dict = task_class.arguments_as_dict(*args, **kwargs)
    name = task_class.generate_name(arguments_dict)
    total_steps = task_class.calculate_total_steps(arguments_dict)
    parent_name = kwargs.get('user_task_name', '')
    chord_data = message_body['chord']
    group_id = message_body['taskset']
    with transaction.atomic():
        group, created = UserTaskStatus.objects.get_or_create(
            task_id=group_id, defaults={'is_container': True, 'name': parent_name, 'task_class': 'celery.group',
                                        'total_steps': total_steps, 'user_id': user_id})
        if created:
            # Also create a status for the chord as a whole
            chord = UserTaskStatus.objects.create(
                is_container=True, name=parent_name, task_class='celery.chord', task_id=str(uuid4()),
                total_steps=total_steps, user_id=user_id)
            group.parent = chord
            group.save(update_fields={'parent', 'modified'})
        else:
            chord = None
            group.increment_total_steps(total_steps)
            if parent_name and not group.name:
                group.set_name(parent_name)
        UserTaskStatus.objects.create(
            name=name, parent=group, task_class=task_class, task_id=task_id, total_steps=total_steps, user_id=user_id)
        # chord body task status
        if not created:
            # body being handled by another of the tasks in the header
            return
        task_id = chord_data['options']['task_id']
        body_task = chord_data['task']
        body_class = import_string(body_task).__class__
        if not issubclass(body_class, UserTaskMixin):
            return
        args = chord_data['args']
        kwargs = chord_data['kwargs']
        arguments_dict = body_class.arguments_as_dict(*args, **kwargs)
        name = body_class.generate_name(arguments_dict)
        total_steps = body_class.calculate_total_steps(arguments_dict)
        UserTaskStatus.objects.get_or_create(
            task_id=task_id, defaults={'name': name, 'parent': chord, 'task_class': body_task,
                                       'total_steps': total_steps, 'user_id': user_id})
        chord.increment_total_steps(total_steps)


def _get_or_create_group_parent(message_body, user_id):
    """
    Determine if the given task belongs to a group or not, and if so, get or create a status record for the group.

    Arguments:
        message_body (dict): The body of the before_task_publish signal for the task in question
        user_id (int): The primary key of the user model record for the user who triggered the task.
                       (If using a custom user model, this may not be an integer.)

    Returns
    -------
        UserTaskStatus: The status record for the containing group, or `None` if there isn't one

    """
    parent_id = message_body.get('taskset', None)
    if not parent_id:
        # Not part of a group
        return None
    parent_class = 'celery.group'
    parent_name = message_body['kwargs'].get('user_task_name', '')
    parent, _ = UserTaskStatus.objects.get_or_create(
        task_id=parent_id, defaults={'is_container': True, 'name': parent_name, 'task_class': parent_class,
                                     'total_steps': 0, 'user_id': user_id})
    if parent_name and not parent.name:
        parent.name = parent_name
        parent.save(update_fields={'name', 'modified'})
    return parent


def _get_user_id(arguments_dict):
    """
    Get and validate the `user_id` argument to a task derived from `UserTaskMixin`.

    Arguments:
        arguments_dict (dict): The parsed positional and keyword arguments to the task

    Returns
    -------
        int: The primary key of a user record (may not be an int if using a custom user model)

    """
    if 'user_id' not in arguments_dict:
        raise TypeError('Each invocation of a UserTaskMixin subclass must include the user_id')
    user_id = arguments_dict['user_id']
    try:
        get_user_model().objects.get(pk=user_id)
    except (ValueError, get_user_model().DoesNotExist):
        raise TypeError('Invalid user_id: {}'.format(user_id))
    return user_id


@task_prerun.connect
def start_user_task(sender=None, **kwargs):  # pylint: disable=unused-argument
    """
    Update the status record when execution of a :py:class:`UserTaskMixin` begins.
    """
    try:
        current_connection = transaction.get_connection()
    except Exception:  # pylint: disable=broad-except
        current_connection = None

    # We may be running this task in the context of an atomic transaction.
    # If so, it's possible that the current transaction could be forced closed
    # due to this block of code, where the `AUTOCOMMIT` defined in settings
    # does not match the autocommit state of the current transaction:
    # https://github.com/django/django/blob/1.11.24/django/db/backends/base/base.py#L510-L512
    # Thus, we should only try to close old connections if we're not currently in an atomic block.
    if current_connection and (not current_connection.in_atomic_block):
        # This should close any obsolete connections,
        # forcing Django to grab a new connection the next time it makes a query.
        # See: https://github.com/django/django/blob/master/django/db/__init__.py#L55
        close_old_connections()

    if isinstance(sender, UserTaskMixin):
        sender.status.start()


@task_failure.connect
def task_failed(sender=None, **kwargs):
    """
    Update the status record accordingly when a :py:class:`UserTaskMixin` fails.
    """
    if isinstance(sender, UserTaskMixin):
        exception = kwargs['exception']
        if not isinstance(exception, TaskCanceledException):
            # Don't include traceback, since this is intended for end users
            sender.status.fail(str(exception))
        user_task_stopped.send_robust(sender=UserTaskStatus, status=sender.status)


@task_retry.connect
def retrying_task(sender=None, **kwargs):  # pylint: disable=unused-argument
    """
    Update the status record to reflect that a retry is pending.
    """
    if isinstance(sender, UserTaskMixin):
        sender.status.retry()


@task_success.connect
def task_succeeded(sender=None, **kwargs):  # pylint: disable=unused-argument
    """
    Update the status record accordingly when a :py:class:`UserTaskMixin` finishes successfully.
    """
    if isinstance(sender, UserTaskMixin):
        status = sender.status
        # Failed tasks with good exception handling did not succeed just because they ended cleanly
        if status.state not in (UserTaskStatus.CANCELED, UserTaskStatus.FAILED, UserTaskStatus.RETRYING):
            status.succeed()
        user_task_stopped.send_robust(sender=UserTaskStatus, status=sender.status)



# -*- coding: utf-8 -*-
"""
Custom exception classes.
"""

from __future__ import absolute_import, unicode_literals


class TaskCanceledException(RuntimeError):
    """
    Raised to stop task execution when a cancellation command is noticed.
    """



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-29 13:03
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('user_tasks', '0002_artifact_file_storage'),
    ]

    operations = [
        migrations.AlterField(
            model_name='usertaskartifact',
            name='url',
            field=models.URLField(blank=True, max_length=512),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Django admin configuration for the ``django-user-tasks`` models.
"""

from __future__ import absolute_import, unicode_literals

from django.contrib import admin

from .models import UserTaskArtifact, UserTaskStatus


class UserTaskArtifactAdmin(admin.ModelAdmin):
    """
    Configuration for UserTaskArtifact admin panel.
    """

    list_display = ('created', 'uuid', 'status', 'name', 'text')
    ordering = ('-created',)
    search_fields = ('uuid', 'name', 'text')


class UserTaskStatusAdmin(admin.ModelAdmin):
    """
    Configuration for UserTaskStatus admin panel.
    """

    list_display = ('created', 'uuid', 'state', 'user', 'name')
    ordering = ('-created',)
    search_fields = (
        'uuid', 'task_id', 'task_class', 'name', 'user__username', 'user__email'
    )


admin.site.register(UserTaskArtifact, UserTaskArtifactAdmin)
admin.site.register(UserTaskStatus, UserTaskStatusAdmin)



# -*- coding: utf-8 -*-
"""
URLs for user_tasks.
"""
from __future__ import absolute_import, unicode_literals

from rest_framework.routers import SimpleRouter

from user_tasks.views import ArtifactViewSet, StatusViewSet

ROUTER = SimpleRouter()
ROUTER.register(r'artifacts', ArtifactViewSet, base_name='usertaskartifact')
ROUTER.register(r'tasks', StatusViewSet, base_name='usertaskstatus')

urlpatterns = ROUTER.urls



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2019-05-29 17:51
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('user_tasks', '0003_url_max_length'),
    ]

    operations = [
        migrations.AlterField(
            model_name='usertaskartifact',
            name='url',
            field=models.TextField(blank=True, validators=[django.core.validators.URLValidator()]),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Custom Django settings for django-user-tasks.
"""

from __future__ import absolute_import, unicode_literals

from datetime import timedelta

from django.conf import settings as django_settings
from django.core.files.storage import get_storage_class

from user_tasks import filters


class LazySettings():
    """
    The behavior of ``django-user-tasks`` can be customized via the following Django settings.
    """

    @property
    def USER_TASKS_ARTIFACT_FILTERS(self):  # pylint: disable=invalid-name
        """
        Tuple containing zero or more filters for UserTaskArtifact listing REST API calls.

        Each entry should be a Django REST Framework filter backend class
        object, such as ``django_filters.rest_framework.DjangoFilterBackend``.
        The default value contains only ``user_tasks.filters.ArtifactFilterBackend``,
        which allows superusers to see all artifacts but other users to see only
        those for artifacts they triggered themselves.
        """
        return getattr(django_settings, 'USER_TASKS_STATUS_FILTERS', (filters.ArtifactFilterBackend,))

    @property
    def USER_TASKS_ARTIFACT_STORAGE(self):  # pylint: disable=invalid-name
        """
        File storage backend to use for :py:attr:`user_tasks.models.UserTaskStatus.file`.

        If explicitly set, the setting should be the import path of a storage
        backend class.
        """
        import_path = getattr(django_settings, 'USER_TASKS_ARTIFACT_STORAGE', None)
        return get_storage_class(import_path)()

    @property
    def USER_TASKS_MAX_AGE(self):  # pylint: disable=invalid-name
        """
        ``timedelta`` reflecting the age after which UserTaskStatus records should be deleted.

        For this setting to be useful, ``user_tasks.tasks.purge_old_user_tasks``
        should be configured to run on an appropriate schedule.  Note that the
        age is calculated from task creation, not completion.  The default
        value is 30 days.
        """
        return getattr(django_settings, 'USER_TASKS_MAX_AGE', timedelta(days=30))

    @property
    def USER_TASKS_STATUS_FILTERS(self):  # pylint: disable=invalid-name
        """
        Tuple containing zero or more filters for UserTaskStatus listing REST API calls.

        Each entry should be a Django REST Framework filter backend class
        object, such as ``django_filters.rest_framework.DjangoFilterBackend``.
        The default value contains only ``user_tasks.filters.StatusFilterBackend``,
        which allows superusers to see all task statuses but other users to see only
        those for tasks they triggered themselves.
        """
        return getattr(django_settings, 'USER_TASKS_STATUS_FILTERS', (filters.StatusFilterBackend,))


settings = LazySettings()  # pylint: disable=invalid-name



# -*- coding: utf-8 -*-
"""
Celery task abstract base classes.
"""

from __future__ import absolute_import, unicode_literals

import inspect
import logging

from celery import shared_task
from celery.task import Task

from django.utils.timezone import now

from .conf import settings
from .models import UserTaskStatus

LOGGER = logging.getLogger(__name__)


class UserTaskMixin():
    """
    Mixin class for user-triggered Celery tasks.

    Subclasses should usually override :py:meth:`generate_name` and
    :py:meth:`calculate_total_steps`.  In order to access the
    :py:attr:`status` property (for calling its
    :py:meth:`~user_tasks.models.UserStatus.increment_completed_steps` and
    :py:meth:`~user_tasks.models.UserStatus.set_state` methods), task
    functions should generally be bound (``bind=True`` in the task decorator)
    and have a ``self`` parameter.

    Additionally, all task functions using a UserTaskMixin subclass must
    provide a ``user_id`` parameter, as either a positional or keyword
    argument.
    """

    @classmethod
    def generate_name(cls, arguments_dict):  # pylint: disable=unused-argument
        """
        Generate a name for the corresponding :py:class:`~user_tasks.models.UserTaskStatus` model instance.

        Should be implemented by each subclass to generate a meaningful name
        from the task parameters.  Defaults to the name of the task function.
        """
        return cls.__name__.split('.')[-1]

    @staticmethod
    def calculate_total_steps(arguments_dict):  # pylint: disable=unused-argument
        """
        Determine from the task's parameters how many total steps the task will perform.

        By default, there is only 1 step (the entire task); to allow for more
        useful progress bars in the UI, override this method to specify a
        meaningful larger number and call ``self.status.increment_completed_steps()``
        periodically during task execution.
        """
        return 1

    @classmethod
    def arguments_as_dict(cls, *args, **kwargs):
        """
        Generate the arguments dictionary provided to :py:meth:`generate_name` and :py:meth:`calculate_total_steps`.

        This makes it possible to fetch arguments by name regardless of
        whether they were passed as positional or keyword arguments.  Unnamed
        positional arguments are provided as a tuple under the key ``pos``.
        """
        all_args = (None, ) + args
        return inspect.getcallargs(cls.run, *all_args, **kwargs)  # pylint: disable=deprecated-method

    @property
    def status(self):
        """
        Get the :py:class:`~user_tasks.models.UserTaskStatus` model instance for this UserTaskMixin.
        """
        task_id = self.request.id
        try:
            # Most calls are for existing objects, don't waste time
            # preparing creation arguments unless necessary
            return UserTaskStatus.objects.get(task_id=task_id)
        except UserTaskStatus.DoesNotExist:
            # Probably an eager task that skipped the before_task_publish
            # signal (or an atomic view where the new record hasn't been
            # committed yet).  Create a record for it.
            arguments_dict = self.arguments_as_dict(*self.request.args, **self.request.kwargs)
            name = self.generate_name(arguments_dict)
            task_class = '.'.join([self.__class__.__module__, self.__class__.__name__])
            total_steps = self.calculate_total_steps(arguments_dict)
            user_id = arguments_dict['user_id']
            # Use get_or_create() again just in case another process created it in the meantime
            return UserTaskStatus.objects.get_or_create(
                task_id=task_id, defaults={'user_id': user_id, 'name': name, 'task_class': task_class,
                                           'total_steps': total_steps})[0]


class UserTask(Task, UserTaskMixin):  # pylint: disable=abstract-method
    """
    Abstract base class for user-triggered Celery tasks.

    See :py:class:`UserTaskMixin` for details on how to implement and use
    subclasses.
    """

    abstract = True


@shared_task
def purge_old_user_tasks():
    """
    Delete any UserTaskStatus and UserTaskArtifact records older than ``settings.USER_TASKS_MAX_AGE``.

    Intended to be run as a scheduled task.
    """
    limit = now() - settings.USER_TASKS_MAX_AGE
    # UserTaskArtifacts will also be removed via deletion cascading
    UserTaskStatus.objects.filter(created__lt=limit).delete()



# -*- coding: utf-8 -*-
"""
Database models for user_tasks.
"""

from __future__ import absolute_import, unicode_literals

import logging
from uuid import uuid4

from celery import current_app

from django.conf import settings as django_settings
from django.core.validators import URLValidator
from django.db import models, transaction
from django.db.models import Q
from django.db.models.expressions import F
from django.utils.encoding import python_2_unicode_compatible
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from model_utils.models import TimeStampedModel

from user_tasks import user_task_stopped

from .conf import settings
from .exceptions import TaskCanceledException

LOGGER = logging.getLogger(__name__)

# The pylint "disable=no-member" comments are needed because of the 'self' argument to the parent field definition.
# See https://github.com/landscapeio/pylint-django/issues/35 for more details


@python_2_unicode_compatible
class UserTaskStatus(TimeStampedModel):
    """
    The current status of an asynchronous task running on behalf of a particular user.

    The methods of this class should generally not be run as part of larger
    transactions.  If they are, some deadlocks and race conditions become
    possible.
    """

    PENDING = 'Pending'
    IN_PROGRESS = 'In Progress'
    SUCCEEDED = 'Succeeded'
    FAILED = 'Failed'
    CANCELED = 'Canceled'
    RETRYING = 'Retrying'

    STATE_TRANSLATIONS = {
        PENDING: _('Pending'),
        IN_PROGRESS: _('In Progress'),
        SUCCEEDED: _('Succeeded'),
        FAILED: _('Failed'),
        CANCELED: _('Canceled'),
        RETRYING: _('Retrying'),
    }

    uuid = models.UUIDField(default=uuid4, unique=True, editable=False, help_text='Unique ID for use in APIs')
    user = models.ForeignKey(
        django_settings.AUTH_USER_MODEL,
        help_text='The user who triggered the task',
        on_delete=models.CASCADE,
    )
    # Auto-generated Celery task IDs will always be 36 characters, but longer custom ones can be specified
    task_id = models.CharField(max_length=128, unique=True, db_index=True,
                               help_text='UUID of the associated Celery task')
    parent = models.ForeignKey('self', on_delete=models.CASCADE, null=True, blank=True, default=None,
                               help_text='Status of the containing task grouping (if any)')
    is_container = models.BooleanField(default=False,
                                       help_text='True if this status corresponds to a container of multiple tasks')
    task_class = models.CharField(max_length=128, help_text='Fully qualified class name of the task being performed')
    name = models.CharField(max_length=255, help_text='A name for this task which the triggering user will understand')
    state = models.CharField(max_length=128, default=PENDING)
    completed_steps = models.PositiveSmallIntegerField(default=0)
    total_steps = models.PositiveSmallIntegerField()
    attempts = models.PositiveSmallIntegerField(default=1, help_text='How many times has execution been attempted?')

    class Meta:
        """
        Additional configuration for the UserTaskStatus model.
        """

        verbose_name_plural = 'user task statuses'

    def start(self):
        """
        Mark the task as having been started (as opposed to waiting for an available worker), and save it.
        """
        if self.state == UserTaskStatus.CANCELED and not self.is_container:
            raise TaskCanceledException
        self.state = UserTaskStatus.IN_PROGRESS
        self.save(update_fields={'state', 'modified'})
        if self.parent_id:
            with transaction.atomic():
                parent = UserTaskStatus.objects.select_for_update().get(pk=self.parent_id)
                if parent.state in (UserTaskStatus.PENDING, UserTaskStatus.RETRYING):
                    parent.start()

    def increment_completed_steps(self, steps=1):
        """
        Increase the value of :py:attr:`completed_steps` by the given number and save, then check for cancellation.

        If cancellation of the task has been requested, a TaskCanceledException
        will be raised to abort execution.  If any special cleanup is required,
        this exception should be caught and handled appropriately.

        This method should be called often enough to provide a useful
        indication of progress, but not so often as to cause undue burden on
        the database.
        """
        UserTaskStatus.objects.filter(pk=self.id).update(completed_steps=F('completed_steps') + steps,
                                                         modified=now())
        self.refresh_from_db(fields={'completed_steps', 'modified', 'state'})
        if self.parent:
            self.parent.increment_completed_steps(steps)  # pylint: disable=no-member
        # Was a cancellation command recently sent?
        if self.state == self.CANCELED and not self.is_container:
            raise TaskCanceledException

    def increment_total_steps(self, steps):
        """Increase the value of :py:attr:`total_steps` by the given number and save."""
        # Assume that other processes may be making concurrent changes
        UserTaskStatus.objects.filter(pk=self.id).update(total_steps=F('total_steps') + steps, modified=now())
        self.refresh_from_db(fields={'total_steps', 'modified'})
        if self.parent:
            self.parent.increment_total_steps(steps)  # pylint: disable=no-member

    def cancel(self):
        """
        Cancel the associated task if it hasn't already finished running.
        """
        if self.is_container:
            children = UserTaskStatus.objects.filter(parent=self)
            for child in children:
                child.cancel()
        elif self.state in (UserTaskStatus.PENDING, UserTaskStatus.RETRYING):
            current_app.control.revoke(self.task_id)
            user_task_stopped.send_robust(UserTaskStatus, status=self)
        with transaction.atomic():
            status = UserTaskStatus.objects.select_for_update().get(pk=self.id)
            if status.state in (UserTaskStatus.CANCELED, UserTaskStatus.FAILED, UserTaskStatus.SUCCEEDED):
                return
            status.state = UserTaskStatus.CANCELED
            status.save(update_fields={'state', 'modified'})
            self.state = status.state
            self.modified = status.modified

    def fail(self, message):
        """
        Mark the task as having failed for the given reason, and save it.

        The message will be available as the :py:attr:`UserTaskArtifact.text`
        field of a UserTaskArtifact with name "Error".  There may be more than
        one such artifact for a single UserTaskStatus, especially if it
        represents a container for multiple parallel tasks.
        """
        with transaction.atomic():
            UserTaskArtifact.objects.create(status=self, name='Error', text=message)
            self.state = UserTaskStatus.FAILED
            self.save(update_fields={'state', 'modified'})
        if self.parent:
            self.parent.fail(message)  # pylint: disable=no-member

    def retry(self):
        """
        Update the status to reflect that a problem was encountered and the task will be retried later.
        """
        # Note that a retry does not affect the state of a containing task
        # grouping; it's effectively still in progress
        self.attempts += 1
        self.state = UserTaskStatus.RETRYING
        self.save(update_fields={'attempts', 'state', 'modified'})

    def set_name(self, name):
        """
        Give the specified name to this status and all of its ancestors.
        """
        self.name = name
        self.save(update_fields={'name', 'modified'})
        if self.parent:
            self.parent.set_name(name)  # pylint: disable=no-member

    def set_state(self, custom_state):
        """
        Set the state to a custom in-progress value.

        This can be done to indicate which stage of a long task is currently
        being executed, like "Sending email messages".
        """
        with transaction.atomic():
            status = UserTaskStatus.objects.select_for_update().get(pk=self.id)
            if status.state == UserTaskStatus.CANCELED and not self.is_container:
                raise TaskCanceledException
            status.state = custom_state
            status.save(update_fields={'state', 'modified'})
            self.state = status.state
            self.modified = status.modified
        if self.parent and self.parent.task_class != 'celery.group':  # pylint: disable=no-member
            self.parent.set_state(custom_state)  # pylint: disable=no-member

    @property
    def state_text(self):
        """
        Get the translation into the current language of the current state of this status instance.
        """
        return self.STATE_TRANSLATIONS.get(self.state, self.state)

    def succeed(self):
        """
        Mark the task as having finished successfully and save it.
        """
        if self.completed_steps < self.total_steps:
            self.increment_completed_steps(self.total_steps - self.completed_steps)
        self.state = UserTaskStatus.SUCCEEDED
        self.save(update_fields={'state', 'modified'})
        if self.parent_id:
            query = UserTaskStatus.objects.filter(~Q(state=UserTaskStatus.SUCCEEDED), parent__pk=self.parent_id)
            if not query.exists():
                self.parent.succeed()  # pylint: disable=no-member

    def __str__(self):
        """
        Get a string representation of this task.
        """
        return '<UserTaskStatus: {}>'.format(self.name)


@python_2_unicode_compatible
class UserTaskArtifact(TimeStampedModel):
    """
    An artifact (or error message) generated for a user by an asynchronous task.

    May be a file, a URL, or text stored directly in the database table.
    """

    uuid = models.UUIDField(default=uuid4, unique=True, editable=False, help_text='Unique ID for use in APIs')
    status = models.ForeignKey(UserTaskStatus, on_delete=models.CASCADE, related_name='artifacts')
    name = models.CharField(max_length=255, default='Output',
                            help_text='Distinguishes between multiple artifact types for the same task')
    file = models.FileField(null=True, blank=True, storage=settings.USER_TASKS_ARTIFACT_STORAGE,
                            upload_to='user_tasks/%Y/%m/%d/')
    url = models.TextField(blank=True, validators=[URLValidator()])
    text = models.TextField(blank=True)

    def __str__(self):
        """
        Get a string representation of this artifact.
        """
        if self.file:
            content = self.file.name
        elif self.url:
            content = self.url
        elif len(self.text) > 50:
            content = self.text[:47] + '...'
        else:
            content = self.text
        return '<UserTaskArtifact: ({}) {}>'.format(self.name, content)



# -*- coding: utf-8 -*-
"""
user_tasks Django application initialization.
"""

from __future__ import absolute_import, unicode_literals

from django.apps import AppConfig


class UserTasksConfig(AppConfig):
    """
    Configuration for the user_tasks Django application.
    """

    name = 'user_tasks'
    verbose_name = 'User Tasks'

    def ready(self):
        """
        Register Celery signal handlers.
        """
        import user_tasks.signals  # pylint: disable=unused-variable, import-outside-toplevel, unused-import
